global K Y prior C tol
delta = 0.01; % confidence on the bound. The bound holds with probability 
            % greater than 1-delta
% Load dataset
load pim.mat
% xm are the training observations, ym are the corresponding labels
% xt and yt serve as test set with analogous roles.

[m, dim] = size(xm); % number of training patterns and length of
                     % the observations
t = length(yt); % number of testing patterns
sigma = sqrt(dim); % spread of the Gaussian kernel: k(x,y) = exp(-0.5*||x-y||^2/sigma/sigma)
C = 10; % SVM regularization 
tol = 1e-3; % initial tolerance for the KKT conditions in the SVM.

% Train the SVM. I use a quick and dirty implementation of Platt's
% SMO that trains an SVM without bias term (b=0). It is 
% recommendable your replacing this implementation with a more
% reliable one.

% First I compute the overall kernel matrix with the train and test
% data;
xall = [xm;xt];
idxtrain = 1:m;
idxtest = m+1:m+t;
Kall = gaussiankernel(xall, xall, sigma);

K = Kall(idxtrain,idxtrain); % only the training kernel matrix goes
                             % to the optimization
Y = ym; 
[w1, pos, obj] = unbiased_svm_globalK();

idsvs = idxtrain(pos); % position of the SVs

% compute the PAC-Bayes bound with a non-informative prior: a
% Spherical Gaussian centred on the origin
pred = Kall(:,idsvs)*w1;
Mod = sqrt(w1'*pred(idsvs)); %norm of the weight vector: w'*K*w,
                             %where w are the signed nonzero
                             %Lagrange multipliers 

% Observed Test Error Rate
TestErrorSVM = mean((yt.*pred(idxtest)) < 0);

fprintf('\n Observed Test Error Rate %.2f\n', TestErrorSVM);
fprintf('--------------------------------------------------\n');
% PAC-BAYES BOUND for SVM
gamma = ym.*pred(idxtrain);
[PACBayesBound, PACmu] = pac_bayes_bound(gamma/Mod, delta);

fprintf('\n PAC Bayes Bound on the true error: %.2f\n', PACBayesBound);
fprintf('--------------------------------------------------\n');
% PRIOR PAC Bayes Bound with the prior being a mixture of J
% spherical Gaussians

% Definition of the prior
J = 10; % number of components in the mixture of Gaussians Prior.
eta_j = linspace(1, 100, J);  % Scalings of the components of the
                            % mixture, each component is a Gaussian
                            % with identity covariance matrix
                            % centred along the weight vector of
                            % the prior at a distance eta_j(j) of
                            % the origin

r = 0.5; % fraction of the training set left out to learn the prior
         % (and thus not used in the computation of the bound)

order = randperm(m)';
Nprior = round(m*r); % number of training instances used to learn
                       % the prior

idxprior = sort(order(1:Nprior));
idxposterior = sort(order(1+Nprior:m));

K = Kall(idxprior, idxprior);
Y = ym(idxprior);
[w2, pos2, obj] = unbiased_svm_globalK();

idsvs2 = idxprior(pos2);
pred2 = Kall(:,idsvs2)*w2;
Mod2 = sqrt(w2'*pred2(idsvs2));
wrw = w1'*pred2(idsvs)/Mod2/Mod; 
[PriorPACBayesBound, best_eta, ...
 PriorPACBayesMu] = prior_pac_bayes_bound(gamma(idxposterior)/Mod, ...
                                          delta, eta_j,wrw);

fprintf('\n Prior PAC Bayes Bound on the true error: %.2f\n', PriorPACBayesBound);
fprintf('--------------------------------------------------\n');

fprintf('\t Analysis of the Prior PAC Bayes Bound with %d priors,\n \t and %.2f fraction of the training data to learn the prior\n ', J, r);
 

% TAU BOUND for SVM
% Now the prior distribution on classifiers is a Gaussian centred
% on a unitary vector wr at a distance eta0 from the origin and with a covariance matrix such that all the variances
% are equal to 1 but along the direction given by the prior, where
% it is equal to tau.

eta0=50;
tau=50;
    [TauBound, TauMu] = taupsvm_bound(gamma(idxposterior)/Mod, delta, tau, eta0, wrw);

    fprintf('\n Tau Prior PAC Bayes Bound on the true error: %.2f\n', TauBound);
    fprintf('--------------------------------------------------\n');
fprintf('\t Analysis of the Tau Prior PAC Bayes Bound with a priors stretched tau=%.2f,\n \t and %.2f fraction of the training data to learn the prior\n \t and centering the prior at a distance %.2f of the origin\n', tau, r, eta0);

        
        