%
% [Q_D, eta_best, mu_best] = prior_pac_bayes_bound(gamma, delta, w, v_eta
% ,wrw) 
%
% Computes the Prior PAC-Bayes bound for a classifier with normalized
% margin gamma.
%
% INPUT:
%
% gamma: Normalized margin of the classifier: 
%         gamma(i) = y(i)* w'* x(i) / (|w| |x(i)|)
% delta: probability that the bound does not hold
% v_eta: vector of lengths of the weight vector of the prior
%        distribution. Values for the search
% wrw: w_r'*w / (|w_r| |w|) where w_r/|w_r| is the unitary prior
%          weight of classifiers and w/|w| is the unitary posterior
%
% OUTPUT:
%
% eta_best: distance from the prior to the origin. It's the best
% component within v_eta
% mu_best: distance from the posterior to the origin
% QD: Expected true error of the classifier
%
% EPH, G2PI/DTSC/UC3M  June 20th, 2006
%

function [QD, eta_best,mu_best]= prior_pac_bayes_bound(gammaw, delta,v_eta, wrw)  
  mr = length(gammaw); % amount of data to calculate the posterior
  size_eta= length(v_eta); % number of priors. 

  ln = log((mr+1)/delta) + log(size_eta); % penalty term for using priors
                                         % in the computation of the bound
    
  %%%%%%%%%%%%%%%%%%%%%%%%%%
  %                        %
  % Search for optimal eta %
  %                        %
  %%%%%%%%%%%%%%%%%%%%%%%%%%
  
  if size_eta==1
    [QD, mu_best] =  pac_bayes_bound(gammaw,delta, 1, v_eta,wrw*v_eta);
    eta_best = v_eta;
    return
  end
  
  % first we compute the bound for four values of the search range
  ee = round(linspace(1,size_eta,4));
  for epi=1:length(ee)
    [bound(epi), mu_opt(epi)] =  pac_bayes_bound(gammaw,delta, size_eta, ...
                                                  v_eta(ee(epi)),wrw*v_eta(ee(epi))); 
  end
  
  done = 0; % done = 1 means we have found best component of the
            % mixture prior
  QD = 200;
  while done == 0
    [pas cual] = min(bound);
    if pas < QD
      QD = pas;
      eta_best = v_eta(ee(cual));
      mu_best = mu_opt(cual);
    end
    
    switch cual
     case 1,
      gap = ee(2) - ee(1);
      if  gap == 1 % we are done
        return 
        
      else
        ee = [ee(1) floor(gap/2) + ee(1) ee(2) ee(3)];
        bound = [bound(1) 100 bound(2:3)];
        mu_opt = [mu_opt(1) 100 mu_opt(2:3)];
        nuevo = 2;
      end
     case 2,
      gap_izq = ee(2)-ee(1);
      gap_der = ee(3) - ee(2);
      if (gap_izq + gap_der)  > 2
        if bound(1) > bound(3)
          if gap_izq > 1
            ee = [ee(1) floor(gap_izq/2)+ee(1) ee(2) ee(3)];
            bound = [bound(1) 100 bound(2:3)];
            mu_opt = [mu_opt(1) 100 mu_opt(2:3)];
            nuevo = 2;
          else
            ee = [ee(1:2) floor(gap_der/2)+ee(2) ee(3)];
            bound = [bound(1:2) 100 bound(3)];
            mu_opt = [mu_opt(1:2) 100 mu_opt(3)];
            nuevo = 3;
          end
        else
          if gap_der > 1
            ee = [ee(1:2) floor(gap_der/2)+ee(2) ee(3)];
            bound = [bound(1:2) 100 bound(3)];
            mu_opt = [mu_opt(1:2) 100 mu_opt(3)];
            nuevo = 3;
          else
            ee = [ee(1) floor(gap_izq/2)+ee(1) ee(2) ee(3)];
            bound = [bound(1) 100 bound(2:3)];
            mu_opt = [mu_opt(1) 100 mu_opt(2:3)];
            nuevo = 2;
          end
        end
      else
          
        return
      end
      case 3,
      gap_izq = ee(3)-ee(2);
      gap_der = ee(4) - ee(3);
      if (gap_izq + gap_der)  > 2
        if bound(2) > bound(4)
          if gap_izq > 1
            ee = [ee(2) floor(gap_izq/2)+ee(2) ee(3:4)];
            bound = [bound(2) 100 bound(3:4)];
            mu_opt = [mu_opt(2) 100 mu_opt(3:4)];
            nuevo = 2;
          else
            ee = [ee(2:3) floor(gap_der/2)+ee(3) ee(4)];
            bound = [bound(2:3) 100 bound(4)];
            mu_opt = [mu_opt(2:3) 100 mu_opt(4)];
            nuevo = 3;
          end
        else
          if gap_der > 1
            ee = [ee(2:3) floor(gap_der/2)+ee(3) ee(4)];
            bound = [bound(2:3) 100 bound(4)];
            mu_opt = [mu_opt(2:3) 100 mu_opt(4)];
            nuevo = 3;
          else
            ee = [ee(2) floor(gap_izq/2)+ee(2) ee(3:4)];
            bound = [bound(2) 100 bound(3:4)];
            mu_opt = [mu_opt(2) 100 mu_opt(3:4)];
            nuevo = 2;
          end
        end
      else
          
        return
      end
     case 4,
      gap = ee(4)-ee(3);
      if gap ==1
          
        return
      else
        ee = [ee(2:3) floor(gap/2)+ee(3) ee(4)];
        bound = [bound(2:3) 100 bound(4)];
        mu_opt = [mu_opt(2:3) 100 mu_opt(4)];
        nuevo = 3;
      end
    end
    [bound(nuevo), mu_opt(nuevo)] = pac_bayes_bound(gammaw,delta, ...
                                                      size_eta, ...
                                                      v_eta(ee(nuevo)),wrw* ...
                                                      v_eta(ee(nuevo)));    
  end
  
   
